{
/**********************************************************/
/*                       IJpeg  V1.9                      */
/* Read and write jpeg file faster than borland jpeg unit */
/*             Use IJL.dll Intel Jpeg Library             */
/*                                                        */
/* Last release 02/01/2003                Sylvain Taufour */
/*                                sylvain.taufour@free.fr */
/**********************************************************/
}

unit IJpeg;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, ijl;

type
  TJPEGScale = (jsFullSize, jsHalf, jsQuarter, jsEighth);
  TJPEGPixelFormat = (jf24Bit, jf8Bit);
  TJPEGQualityRange = 1..100;
  TJPEGPerformance = (jpBestQuality, jpBestSpeed);

  TIJpeg = class(TGraphic)
  private
    { Private declarations }
    FScale: TJPEGScale;
    FBitmap: TBitmap;
    FPixelFormat: TJPEGPixelFormat;
    FCompressionQuality: TJPEGQualityRange;
    FProgressiveEncoding: Boolean;
    FProgressiveDisplay: Boolean;
    FPerformance: TJPEGPerformance;
    FComment: string;
    JpgFileName: string;
    JpgStream: TStream;
    BFromFile: Boolean;
    BFromStream: Boolean;
    FDecodeBeforePaint: Boolean;
    Decoded: Boolean;
    procedure Decode;
    procedure SetScale(const Value: TJPEGScale);
    procedure GetScaleSize(var iWidth: Integer; var iHeight: Integer; S: TJPEGScale);
  protected
    { Protected declarations }
  public
    { Public declarations }
    property CompressionQuality: TJPEGQualityRange read FCompressionQuality write FCompressionQuality;
    property Comment: string read FComment write FComment;
    property DecodeBeforePaint: Boolean read FDecodeBeforePaint write FDecodeBeforePaint;
    property Scale: TJPEGScale read FScale write SetScale;
    property Performance: TJPEGPerformance read FPerformance write FPerformance;
    property PixelFormat: TJPEGPixelFormat read FPixelFormat write FPixelFormat;
    property ProgressiveDisplay: Boolean read FProgressiveDisplay write FProgressiveDisplay; // just for compatibility with TJpegImage
    property ProgressiveEncoding: Boolean read FProgressiveEncoding write FProgressiveEncoding;
    procedure Assign(Source: TPersistent); override;
    procedure AssignTo(Dest: TPersistent); override;
    constructor Create; override;
    destructor Destroy; override;
    procedure Draw(ACanvas: TCanvas; const ARect: TRect); override;
    function Equals(Graphic: TGraphic): Boolean; override;
    function GetEmpty: Boolean; override;
    function GetHeight: Integer; override;
    function GetPalette: HPALETTE; override;
    function GetTransparent: Boolean; override;
    function GetWidth: Integer; override;
    procedure LoadFromClipboardFormat(AFormat: Word; AData: THandle; APalette: HPALETTE); override;
    procedure LoadFromFile(const FileName: string); override;
    procedure LoadFromResourceID(Instance: THandle; ResID: Integer; ResType: PChar);
    procedure LoadFromResourceName(Instance: THandle; const ResName: string; ResType: PChar);
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToClipboardFormat(var AFormat: Word; var AData: THandle; var APalette: HPALETTE); override;
    procedure SaveToFile(const FileName: string); override;
    procedure SaveToStream(Stream: TStream); override;
    procedure SetHeight(Value: Integer); override;
    procedure SetWidth(Value: Integer); override;
    procedure WriteData(Stream: TStream); override;
  published
    { Published declarations }
  end;

  TJpegImage = TIJpeg;

implementation

{ TIJpeg }

procedure TIJpeg.Assign(Source: TPersistent);
var
  TmpJpg: TIJpeg;
begin
  Decoded := True;
  
  if Source = nil then
    FreeAndNil(FBitmap)
  else if (Source is TIJpeg) and (Source <> Self) then
  begin
    FreeAndNil(FBitmap);
    FBitmap := TBitmap.Create;

    TmpJpg := TIJpeg(Source);
    FBitmap.Assign(TmpJpg.FBitmap);
    FDecodeBeforePaint := TmpJpg.DecodeBeforePaint;
    BFromFile := TmpJpg.BFromFile;
    BFromStream := TmpJpg.BFromStream;
    JpgFileName := TmpJpg.JpgFileName;

    if Assigned(TmpJpg.JpgStream) then
      JpgStream.CopyFrom(TmpJpg.JpgStream, TmpJpg.JpgStream.Size);

    FScale := TmpJpg.Scale;
    Decoded := TmpJpg.Decoded;
    Width := TmpJpg.Width;
    Height := TmpJpg.Height;
    FComment := TmpJpg.Comment;
    FCompressionQuality := TmpJpg.CompressionQuality;
    FPixelFormat := TmpJpg.PixelFormat;
    FProgressiveEncoding := TmpJpg.ProgressiveEncoding;
    FProgressiveDisplay := TmpJpg.ProgressiveDisplay;
    FPerformance := TmpJpg.Performance;
  end
  else if Source is TBitmap then
  begin
    FreeAndNil(FBitmap);
    FBitmap := TBitmap.Create;
    FBitmap.Assign(TBitmap(Source));
  end
  else
    inherited Assign(Source);
end;

procedure TIJpeg.AssignTo(Dest: TPersistent);
begin
  if Dest is TIJpeg then
    Dest.Assign(Self)
  else if Dest is TGraphic then
  begin
    if not Decoded then
      Decode;
    if Empty then
      Dest.Assign(nil)
    else
      (Dest as TGraphic).Assign(FBitmap);
  end
  else
    inherited AssignTo(Dest);
end;

constructor TIJpeg.Create;
begin
  inherited Create;
  
  FBitmap := TBitmap.Create;
  with FBitmap do
  begin
    Width := 0;
    Height := 0;
    Palette := 0;
    PixelFormat := pf24bit;
  end;

  FCompressionQuality := 100;
  FComment := '';
  BFromFile := False;
  BFromStream := False;
  JpgFileName := '';
  FDecodeBeforePaint := True;
  //FDecodeBeforePaint := False;
  Decoded := False;
end;

procedure TIJpeg.Decode;
var
  iWidth, iHeight: Integer;
  iStatus: integer;
  jcprops: TJPEG_CORE_PROPERTIES;
  DIB: TDIBSection;
  Buff: String;
begin
  if not Decoded then
  begin
    // Initialise the JPEG library
    FillChar (jcprops, SizeOf (jcprops), 0);  // Just to be sure...
    iStatus := ijlInit (@jcprops);

    if iStatus = IJL_OK then
    try
      SetLength(FComment, 1024);
      jcprops.jprops.jpeg_comment := PChar(FComment);
      jcprops.jprops.jpeg_comment_size := 1024;

      //***** Parametres pour dcodage ******************
      if BFromFile then
        jcprops.JPGFile := PChar (JpgFileName);

      if BFromStream then
      begin
        SetLength(Buff, JpgStream.Size);
        JpgStream.Read(Buff[1], JpgStream.Size);

        with jcprops do
        begin
          JPGFile := nil;
          JPGBytes := PByte(Buff);
          JPGSizeBytes := JpgStream.Size;
        end;
      end;
      //*************************************************

      iStatus := ijlRead (@jcprops, IJL_JFILE_READPARAMS);
      if iStatus = IJL_OK then
        begin
        iWidth := jcprops.JPGWidth;
        iHeight := jcprops.JPGHeight;
        GetScaleSize(iWidth, iHeight, FScale);

        case jcprops.JPGChannels of
          1:
            begin
              jcprops.JPGColor    := IJL_G;
              jcprops.DIBChannels := 3;
              jcprops.DIBColor    := IJL_BGR;
            end;

          3:
            begin
              jcprops.JPGColor    := IJL_YCBCR;
              jcprops.DIBChannels := 3;
              jcprops.DIBColor    := IJL_BGR;
            end;

          4:
            begin
              jcprops.JPGColor    := IJL_YCBCRA_FPX;
              jcprops.DIBChannels := 4;
              jcprops.DIBColor    := IJL_RGBA_FPX;
            end;
          else
            begin
              jcprops.DIBColor := TIJL_COLOR(IJL_OTHER);
              jcprops.JPGColor := TIJL_COLOR(IJL_OTHER);
              jcprops.DIBChannels := jcprops.JPGChannels;
            end;
        end;

        with FBitmap do
        begin
          if jcprops.DIBChannels = 1 then
            PixelFormat := pf8Bit
          else
            PixelFormat := pf24Bit;

          Width := iWidth;
          Height := iHeight;
        end;

        Width := iWidth;
        Height := iHeight;

        FillChar (DIB, SizeOf (DIB), 0);
        iStatus := GetObject (FBitmap.Handle, SizeOf (DIB), @DIB);
        if iStatus <> 0 then
        begin
          with jcprops do
          begin
            DIBWidth := iWidth;
            DIBHeight := -iHeight;
            DIBPadBytes := IJL_DIB_PAD_BYTES(jcprops.DIBWidth, jcprops.DIBChannels);
            DIBBytes := PByte (DIB.dsBm.bmBits);
          end;

          if BFromFile then
          begin
            case FScale of
              jsHalf: iStatus := ijlRead (@jcprops, IJL_JFILE_READONEHALF);
              jsQuarter: iStatus := ijlRead (@jcprops, IJL_JFILE_READONEQUARTER);
              jsEighth: iStatus := ijlRead (@jcprops, IJL_JFILE_READONEEIGHTH);
            else
              iStatus := ijlRead (@jcprops, IJL_JFILE_READWHOLEIMAGE);
            end;
          end;

          if BFromStream then
          begin
            case FScale of
              jsHalf: iStatus := ijlRead (@jcprops, IJL_JBUFF_READONEHALF);
              jsQuarter: iStatus := ijlRead (@jcprops, IJL_JBUFF_READONEQUARTER);
              jsEighth: iStatus := ijlRead (@jcprops, IJL_JBUFF_READONEEIGHTH);
            else
              iStatus := ijlRead (@jcprops, IJL_JBUFF_READWHOLEIMAGE);
            end;
          end;

          if iStatus >= 0 then
            FBitmap.Modified := True;

          SetLength(FComment, Length(FComment));
        end;
      end;

      if iStatus <> IJL_OK then
      begin
        if BFromFile then
          raise (EReadError.Create ('Error reading JPEG from file'));

        if BFromStream then
          raise (EReadError.Create ('Error reading JPEG from stream'));
      end;

    finally
      ijlFree (@jcprops);
    end;
  end;
  
  Decoded := True;
  BFromFile := False;
  BFromStream := False;
  JpgFileName := '';
  if BFromStream then
    JpgStream.Free;
end;

destructor TIJpeg.Destroy;
begin
  if FBitmap <> nil then
  begin

    if FBitmap.Palette <> 0 then
      DeleteObject(FBitmap.Palette);
    FBitmap.Free;
  end;

  inherited Destroy;
end;

procedure TIJpeg.Draw(ACanvas: TCanvas; const ARect: TRect);
begin
  if not Decoded and FDecodeBeforePaint then
    Decode;

  if FBitmap <> nil then
    ACanvas.StretchDraw(ARect, FBitmap);
end;

function TIJpeg.Equals(Graphic: TGraphic): Boolean;
begin
  if FBitmap = nil then
    Result := Graphic = nil
  else
    Result := (Graphic is TIJpeg) and (FBitmap = TIJpeg(TIJpeg).FBitmap);
end;

function TIJpeg.GetEmpty: Boolean;
begin
  Result := FBitmap = nil;
end;

function TIJpeg.GetHeight: Integer;
begin
  if FBitmap <> nil then
    Result := FBitmap.Height
  else
    Result := -1;
end;

function TIJpeg.GetPalette: HPALETTE;
begin
  if FBitmap <> nil then
    Result := FBitmap.Palette
  else
    Result := 0;
end;

function TIJpeg.GetTransparent: Boolean;
begin
  if FBitmap <> nil then
    Result := FBitmap.Transparent
  else
    Result := False;
end;

function TIJpeg.GetWidth: Integer;
begin
  if FBitmap <> nil then
    Result := FBitmap.Width
  else
    Result := -1;
end;

procedure TIJpeg.LoadFromClipboardFormat(AFormat: Word; AData: THandle;
  APalette: HPALETTE);
begin
  if FBitmap <> nil then
    FBitmap.LoadFromClipboardFormat(AFormat, AData, APalette);
end;

procedure TIJpeg.LoadFromFile(const FileName: string);
var
  iWidth, iHeight: Integer;
  iStatus: integer;
  jcprops: TJPEG_CORE_PROPERTIES;
begin
  JpgFileName := FileName;
  BFromFile := True;
  BFromStream := False;

  if FDecodeBeforePaint then
  begin
    // Initialise the JPEG library
    FillChar (jcprops, SizeOf (jcprops), 0);  // Just to be sure...
    iStatus := ijlInit (@jcprops);

    if iStatus = IJL_OK then
    try
      SetLength(FComment, 1024);
      jcprops.jprops.jpeg_comment := PChar(FComment);
      jcprops.jprops.jpeg_comment_size := 1024;

      jcprops.JPGFile := PChar(JpgFileName);
      iStatus := ijlRead (@jcprops, IJL_JFILE_READPARAMS);
      if iStatus = IJL_OK then
      begin
        iWidth := jcprops.JPGWidth;
        iHeight := jcprops.JPGHeight;
        GetScaleSize(iWidth, iHeight, FScale);

        Width := iWidth;
        Height := iHeight;

        SetLength(FComment, Length(FComment));
      end;

    finally
      ijlFree (@jcprops);
    end;
  end
  else
    Decode;
end;

procedure TIJpeg.LoadFromResourceID(Instance: THandle; ResID: Integer;
  ResType: PChar);
var
  Stream: TStream;
begin
  Stream := TResourceStream.CreateFromID(Instance, ResId, ResType);
  Self.LoadFromStream(Stream);
  Stream.Free;
end;

procedure TIJpeg.LoadFromResourceName(Instance: THandle;
  const ResName: string; ResType: PChar);
var
  Stream: TStream;
begin
  Stream := TResourceStream.Create(Instance, ResName, ResType);
  Self.LoadFromStream(Stream);
  Stream.Free;
end;

procedure TIJpeg.LoadFromStream(Stream: TStream);
var
  iWidth, iHeight: Integer;
  iStatus: integer;
  jcprops: TJPEG_CORE_PROPERTIES;
  Buff: String;
begin
  BFromStream := True;
  BFromFile := False;
  JpgStream := TStream.Create;

  if FDecodeBeforePaint then
    begin
    // Initialise the JPEG library
    FillChar (jcprops, SizeOf (jcprops), 0);  // Just to be sure...
    iStatus := ijlInit (@jcprops);

    if iStatus = IJL_OK then
    try
      SetLength(Buff, Stream.Size);
      Stream.Read(Buff[1], Stream.Size);

      with jcprops do
      begin
        JPGFile := nil;
        JPGBytes := PByte(Buff);
        JPGSizeBytes := Stream.Size;
      end;

      jcprops.jprops.jpeg_comment := PChar(FComment);
      jcprops.jprops.jpeg_comment_size := 1024;

      iStatus := ijlRead(@jcprops, IJL_JBUFF_READPARAMS);
      if iStatus = IJL_OK then
      begin
        iWidth := jcprops.JPGWidth;
        iHeight := jcprops.JPGHeight;
        GetScaleSize(iWidth, iHeight, FScale);

        Width := iWidth;
        Height := iHeight;

        SetLength(FComment, Length(FComment));
      end;
    finally
      ijlFree (@jcprops);
    end;
  end
  else
    Decode;
end;

procedure TIJpeg.SaveToClipboardFormat(var AFormat: Word;
  var AData: THandle; var APalette: HPALETTE);
begin
  if FBitmap <> nil then
    FBitmap.SaveToClipboardFormat(AFormat, AData, APalette);
end;

procedure TIJpeg.SaveToFile(const FileName: string);
var
  jcprops : TJPEG_CORE_PROPERTIES;
  iWidth, iHeight, iNChannels : Integer;
  iStatus, err: integer;
  DIB: TDIBSection;
  B: Boolean;
begin
  if not Decoded then
    Decode;

  B := True;
  FBitmap.PixelFormat := pf24bit;
  // Initialise the JPEG library
  FillChar (jcprops, SizeOf (jcprops), 0);  // Just to be sure...
  iStatus := ijlInit (@jcprops);

  if iStatus = IJL_OK then
  try
    FillChar (DIB, SizeOf(DIB), 0);
    GetObject (FBitmap.Handle, SizeOf (DIB), @DIB);
    iWidth  := DIB.dsBm.bmWidth;
    iHeight := DIB.dsBm.bmHeight;
    case FPixelFormat of
      jf8bit: iNChannels := 1;
      jf24bit: iNChannels := 3;
    else
      Raise EInvalidOperation.Create ('Cannot save bitmap as JPEG with specified PixelFormat');
    end;

    with jcprops do
    begin
      DIBWidth := iWidth;
      DIBHeight := -iHeight;
      DIBChannels := iNChannels;
      case FPixelFormat of
        jf8bit: DIBColor := IJL_G;
        jf24bit: DIBColor := IJL_BGR;
      end;
      DIBPadBytes := IJL_DIB_PAD_BYTES(jcprops.DIBWidth,jcprops.DIBChannels);
      DIBBytes := PByte (DIB.dsBm.bmBits);

      JPGFile := PChar (FileName);
      JPGWidth := iWidth;
      JPGHeight := iHeight;
      JPGChannels := 3;
      JPGColor := IJL_YCBCR;
      jquality := CompressionQuality;

      jprops.jpeg_comment_size := Length(FComment) + 1;
      jprops.jpeg_comment := PChar(FComment);

      if FProgressiveEncoding then
        jprops.progressive_found := 1;
    end;

    err := ijlWrite (@jcprops, IJL_JFILE_WRITEWHOLEIMAGE);
    B := IJL_OK = err;

    ijlFree (@jcprops);
  except
  end;
  
  if not B then
    raise (EWriteError.Create ('Error writing JPEG to file. ' + ijlErrorStr(Err)));
end;

procedure TIJpeg.SaveToStream(Stream: TStream);
var
  jcprops : TJPEG_CORE_PROPERTIES;
  iWidth, iHeight, iNChannels : Integer;
  iStatus: Integer;
  B: Boolean;
  Buff: String;
  BufSize: Integer;
  Err: Integer;
  DIB: TDIBSection;
begin
  if not Decoded then
    Decode;
    
  B := True;
  
  FBitmap.PixelFormat := pf24bit;
  // Initialise the JPEG library
  FillChar (jcprops, SizeOf (jcprops), 0);  // Just to be sure...
  iStatus := ijlInit (@jcprops);

  if iStatus = IJL_OK then
  try
    FillChar (DIB, SizeOf(DIB), 0);
    GetObject (FBitmap.Handle, SizeOf (DIB), @DIB);
    iWidth  := DIB.dsBm.bmWidth;
    iHeight := DIB.dsBm.bmHeight;
    
    BufSize := iWidth * iHeight * 3;
    SetLength(Buff, BufSize);

    case FPixelFormat of
      jf8bit: iNChannels := 1;
      jf24bit: iNChannels := 3;
    else
      Raise EInvalidOperation.Create ('Cannot save bitmap as JPEG with specified PixelFormat');
    end;

    with jcprops do
    begin
      DIBWidth := iWidth;
      DIBHeight := -iHeight;
      DIBBytes := PByte (DIB.dsBm.bmBits);;
      DIBChannels := iNChannels;
      case FPixelFormat of
        jf8bit: DIBColor := IJL_G;
        jf24bit: DIBColor := IJL_BGR;
      end;
      DIBPadBytes := IJL_DIB_PAD_BYTES(jcprops.DIBWidth, jcprops.DIBChannels);

      JPGWidth := iWidth;
      JPGHeight := iHeight;
      JPGFile := nil;
      JPGBytes := PByte(Buff);
      JPGSizeBytes := BufSize;
      JPGChannels := 3;
      JPGColor := IJL_YCBCR;
      //JPGSubsampling := IJL_411;
      jquality := CompressionQuality;

      jprops.jpeg_comment_size := Length(FComment) + 1;
      jprops.jpeg_comment := PChar(FComment);

      if FProgressiveEncoding then
        jprops.progressive_found := 1;
    end;

    Err := ijlWrite (@jcprops, IJL_JBUFF_WRITEWHOLEIMAGE);
    B := IJL_OK = Err;

    if B then
      Stream.Write(Buff[1], BufSize);

    ijlFree (@jcprops);
  except
  end;

  if not B then
    raise (EWriteError.Create ('Error writing JPEG to stream. ' + ijlErrorStr(Err)));
end;

procedure TIJpeg.SetHeight(Value: Integer);
begin
  if FBitmap <> nil then
  begin
    FBitmap.Height := Value;
    Changed(Self);
  end;
end;

procedure TIJpeg.SetScale(const Value: TJPEGScale);
var
  tmpWidth, tmpHeight: Integer;
begin
  FScale := Value;
  if FDecodeBeforePaint then
  begin
    tmpWidth := Width;
    tmpHeight := Height;
    GetScaleSize(tmpWidth, tmpHeight, FScale);
    Width := tmpWidth;
    Height := tmpHeight;
  end;
end;

procedure TIJpeg.SetWidth(Value: Integer);
begin
  if FBitmap <> nil then
  begin
    FBitmap.Width := Value;
    Changed(Self);
  end;
end;

procedure TIJpeg.WriteData(Stream: TStream);
begin
  SaveToStream(Stream);
end;

procedure TIJpeg.GetScaleSize(var iWidth: Integer; var iHeight: Integer; S: TJPEGScale);
begin
  case S of
    jsHalf: begin
             iWidth := (iWidth + 1) shr 1;
             iHeight := (iHeight + 1) shr 1;
             end;
    jsQuarter: begin
             iWidth := (iWidth + 3) shr 2;
             iHeight := (iHeight + 3) shr 2;
             end;
    jsEighth: begin
             iWidth := (iWidth + 7) shr 3;
             iHeight := (iHeight + 7) shr 3;
             end;
  end;
end;

{**************************************************}

initialization
  RegisterClass(TIJpeg);
  TPicture.RegisterFileFormat('jpg', 'JPEG Image', TIJpeg);

finalization
  TPicture.UnRegisterGraphicClass(TIJpeg);

end.
